package ru.dbotthepony.kstarbound

import com.google.gson.JsonElement
import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.api.IStarboundFile
import ru.dbotthepony.kstarbound.defs.player.RecipeDefinition
import ru.dbotthepony.kstarbound.util.KOptional
import ru.dbotthepony.kstarbound.util.ParallelPerform
import java.util.*
import java.util.concurrent.ForkJoinPool
import java.util.concurrent.ForkJoinTask
import java.util.concurrent.locks.ReentrantLock
import kotlin.collections.ArrayList
import kotlin.concurrent.withLock

object RecipeRegistry {
	private val LOGGER = LogManager.getLogger()

	data class Entry(val value: RecipeDefinition, val json: JsonElement, val file: IStarboundFile)

	private val recipesInternal = ArrayList<Entry>()
	private val group2recipesInternal = Object2ObjectOpenHashMap<String, ArrayList<Entry>>()
	private val group2recipesBacking = Object2ObjectOpenHashMap<String, List<Entry>>()
	private val output2recipesInternal = Object2ObjectOpenHashMap<String, ArrayList<Entry>>()
	private val output2recipesBacking = Object2ObjectOpenHashMap<String, List<Entry>>()
	private val input2recipesInternal = Object2ObjectOpenHashMap<String, ArrayList<Entry>>()
	private val input2recipesBacking = Object2ObjectOpenHashMap<String, List<Entry>>()

	val recipes: List<Entry> = Collections.unmodifiableList(recipesInternal)
	val group2recipes: Map<String, List<Entry>> = Collections.unmodifiableMap(group2recipesBacking)
	val output2recipes: Map<String, List<Entry>> = Collections.unmodifiableMap(output2recipesBacking)
	val input2recipes: Map<String, List<Entry>> = Collections.unmodifiableMap(input2recipesBacking)

	private val lock = ReentrantLock()

	fun add(recipe: Entry) {
		lock.withLock {
			val value = recipe.value
			recipesInternal.add(recipe)

			for (group in value.groups) {
				group2recipesInternal.computeIfAbsent(group, Object2ObjectFunction { p ->
					ArrayList<Entry>(1).also {
						group2recipesBacking[p as String] = Collections.unmodifiableList(it)
					}
				}).add(recipe)
			}

			output2recipesInternal.computeIfAbsent(value.output.item.key.left(), Object2ObjectFunction { p ->
				ArrayList<Entry>(1).also {
					output2recipesBacking[p as String] = Collections.unmodifiableList(it)
				}
			}).add(recipe)

			for (input in value.input) {
				input2recipesInternal.computeIfAbsent(input.item.key.left(), Object2ObjectFunction { p ->
					ArrayList<Entry>(1).also {
						input2recipesBacking[p as String] = Collections.unmodifiableList(it)
					}
				}).add(recipe)
			}
		}
	}

	fun load(log: ILoadingLog, fileTree: Map<String, List<IStarboundFile>>, executor: ForkJoinPool): List<ForkJoinTask<*>> {
		val files = fileTree["recipe"] ?: return emptyList()

		val elements = Starbound.gson.getAdapter(JsonElement::class.java)
		val recipes = Starbound.gson.getAdapter(RecipeDefinition::class.java)

		return listOf(executor.submit {
			val line = log.line("Loading recipes...")
			val time = System.nanoTime()
			line.maxElements = files.size

			files.batch(executor) { listedFile ->
				try {
					line.text = ("Loading $listedFile")
					val json = elements.read(listedFile.jsonReader())
					val value = recipes.fromJsonTree(json)
					KOptional.of(Entry(value, json, listedFile))
				} catch (err: Throwable) {
					LOGGER.error("Loading recipe definition file $listedFile", err)
					KOptional.empty()
				} finally {
					line.elements.incrementAndGet()
				}
			}.forEach { add(it) }

			line.text = "Loaded recipes in ${((System.nanoTime() - time) / 1_000_000.0).toLong()}ms"
		})
	}
}
