package ru.dbotthepony.kstarbound

import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.internal.bind.JsonTreeReader
import com.google.gson.stream.JsonReader
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.api.IStarboundFile
import ru.dbotthepony.kstarbound.defs.Json2Function
import ru.dbotthepony.kstarbound.defs.JsonFunction
import ru.dbotthepony.kstarbound.defs.Species
import ru.dbotthepony.kstarbound.defs.StatusEffectDefinition
import ru.dbotthepony.kstarbound.defs.item.TreasurePoolDefinition
import ru.dbotthepony.kstarbound.defs.item.api.IItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.BackArmorItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.ChestArmorItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.CurrencyItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.FlashlightDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.HarvestingToolPrototype
import ru.dbotthepony.kstarbound.defs.item.impl.HeadArmorItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.ItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.LegsArmorItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.LiquidItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.MaterialItemDefinition
import ru.dbotthepony.kstarbound.defs.monster.MonsterSkillDefinition
import ru.dbotthepony.kstarbound.defs.monster.MonsterTypeDefinition
import ru.dbotthepony.kstarbound.defs.npc.NpcTypeDefinition
import ru.dbotthepony.kstarbound.defs.npc.TenantDefinition
import ru.dbotthepony.kstarbound.defs.`object`.ObjectDefinition
import ru.dbotthepony.kstarbound.defs.particle.ParticleDefinition
import ru.dbotthepony.kstarbound.defs.player.TechDefinition
import ru.dbotthepony.kstarbound.defs.projectile.ProjectileDefinition
import ru.dbotthepony.kstarbound.defs.quest.QuestTemplate
import ru.dbotthepony.kstarbound.defs.tile.LiquidDefinition
import ru.dbotthepony.kstarbound.defs.tile.MaterialModifier
import ru.dbotthepony.kstarbound.defs.tile.TileDefinition
import ru.dbotthepony.kstarbound.util.AssetPathStack
import ru.dbotthepony.kstarbound.util.KOptional
import ru.dbotthepony.kstarbound.util.ParallelPerform
import java.util.concurrent.ForkJoinPool
import java.util.concurrent.ForkJoinTask

object Registries {
	private val LOGGER = LogManager.getLogger()

	val tiles = Registry<TileDefinition>("tiles")
	val tileModifiers = Registry<MaterialModifier>("tile modifiers")
	val liquid = Registry<LiquidDefinition>("liquid")
	val species = Registry<Species>("species")
	val statusEffects = Registry<StatusEffectDefinition>("status effects")
	val particles = Registry<ParticleDefinition>("particles")
	val items = Registry<IItemDefinition>("items")
	val questTemplates = Registry<QuestTemplate>("quest templates")
	val techs = Registry<TechDefinition>("techs")
	val jsonFunctions = Registry<JsonFunction>("json functions")
	val json2Functions = Registry<Json2Function>("json 2functions")
	val npcTypes = Registry<NpcTypeDefinition>("npc types")
	val projectiles = Registry<ProjectileDefinition>("projectiles")
	val tenants = Registry<TenantDefinition>("tenants")
	val treasurePools = Registry<TreasurePoolDefinition>("treasure pools")
	val monsterSkills = Registry<MonsterSkillDefinition>("monster skills")
	val monsterTypes = Registry<MonsterTypeDefinition>("monster types")
	val worldObjects = Registry<ObjectDefinition>("objects")

	private fun <T> key(mapper: (T) -> String): (T) -> Pair<String, Int?> {
		return { mapper.invoke(it) to null }
	}

	private fun <T> key(mapper: (T) -> String, mapperInt: (T) -> Int): (T) -> Pair<String, Int> {
		return { mapper.invoke(it) to mapperInt.invoke(it) }
	}

	fun validate(): Boolean {
		var any = false

		any = !tiles.validate() || any
		any = !tileModifiers.validate() || any
		any = !liquid.validate() || any
		any = !species.validate() || any
		any = !statusEffects.validate() || any
		any = !particles.validate() || any
		any = !items.validate() || any
		any = !questTemplates.validate() || any
		any = !techs.validate() || any
		any = !jsonFunctions.validate() || any
		any = !json2Functions.validate() || any
		any = !npcTypes.validate() || any
		any = !projectiles.validate() || any
		any = !tenants.validate() || any
		any = !treasurePools.validate() || any
		any = !monsterSkills.validate() || any
		any = !monsterTypes.validate() || any
		any = !worldObjects.validate() || any

		return !any
	}

	private fun loadStage(
		log: ILoadingLog,
		loader: (ILoadingLog.ILine) -> Unit,
		name: String,
	) {
		if (Starbound.terminateLoading)
			return

		val time = System.currentTimeMillis()
		val line = log.line("Loading $name...".also(LOGGER::info))
		loader(line)
		line.text = ("Loaded $name in ${System.currentTimeMillis() - time}ms".also(LOGGER::info))
	}

	private inline fun <reified T : Any> loadStage(
		log: ILoadingLog,
		executor: ForkJoinPool,
		registry: Registry<T>,
		files: List<IStarboundFile>,
		noinline keyProvider: (T) -> Pair<String, Int?>,
		name: String = registry.name
	) {
		val adapter = Starbound.gson.getAdapter(T::class.java)
		val elementAdapter = Starbound.gson.getAdapter(JsonElement::class.java)

		loadStage(log, loader = {
			it.maxElements = files.size

			files.batch(executor) { listedFile ->
				try {
					it.text = "Loading $listedFile"

					AssetPathStack(listedFile.computeDirectory()) {
						val elem = elementAdapter.read(listedFile.jsonReader())
						val read = adapter.fromJsonTree(elem)
						val keys = keyProvider(read);

						KOptional.of {
							try {
								if (keys.second != null)
									registry.add(keys.first, keys.second!!, read, elem, listedFile)
								else
									registry.add(keys.first, read, elem, listedFile)
							} catch (err: Throwable) {
								LOGGER.error("Loading ${registry.name} definition file $listedFile", err);
							}
						}
					}
				} catch (err: Throwable) {
					LOGGER.error("Loading ${registry.name} definition file $listedFile", err);
					KOptional.empty()
				} finally {
					it.elements.incrementAndGet()
				}
			}.forEach { it.invoke() }
		}, name)
	}

	fun load(log: ILoadingLog, fileTree: Map<String, List<IStarboundFile>>, executor: ForkJoinPool): List<ForkJoinTask<*>> {
		val tasks = ArrayList<ForkJoinTask<*>>()

		tasks.addAll(loadItemDefinitions(log, fileTree, executor))

		tasks.add(executor.submit { loadStage(log, { loadJsonFunctions(it, fileTree["functions"] ?: listOf()) }, "json functions") })
		tasks.add(executor.submit { loadStage(log, { loadJson2Functions(it, fileTree["2functions"] ?: listOf()) }, "json 2functions") })
		tasks.add(executor.submit { loadStage(log, { loadTreasurePools(it, fileTree["treasurepools"] ?: listOf()) }, "treasure pools") })

		tasks.add(executor.submit { loadStage(log, executor, tiles, fileTree["material"] ?: listOf(), key(TileDefinition::materialName, TileDefinition::materialId)) })
		tasks.add(executor.submit { loadStage(log, executor, tileModifiers, fileTree["matmod"] ?: listOf(), key(MaterialModifier::modName, MaterialModifier::modId)) })
		tasks.add(executor.submit { loadStage(log, executor, liquid, fileTree["liquid"] ?: listOf(), key(LiquidDefinition::name, LiquidDefinition::liquidId)) })

		tasks.add(executor.submit { loadStage(log, executor, worldObjects, fileTree["object"] ?: listOf(), key(ObjectDefinition::objectName)) })
		tasks.add(executor.submit { loadStage(log, executor, statusEffects, fileTree["statuseffect"] ?: listOf(), key(StatusEffectDefinition::name)) })
		tasks.add(executor.submit { loadStage(log, executor, species, fileTree["species"] ?: listOf(), key(Species::kind)) })
		tasks.add(executor.submit { loadStage(log, executor, particles, fileTree["particle"] ?: listOf(), key(ParticleDefinition::kind)) })
		tasks.add(executor.submit { loadStage(log, executor, questTemplates, fileTree["questtemplate"] ?: listOf(), key(QuestTemplate::id)) })
		tasks.add(executor.submit { loadStage(log, executor, techs, fileTree["tech"] ?: listOf(), key(TechDefinition::name)) })
		tasks.add(executor.submit { loadStage(log, executor, npcTypes, fileTree["npctype"] ?: listOf(), key(NpcTypeDefinition::type)) })
		// tasks.add(executor.submit { loadStage(log, executor, projectiles, ext2files["projectile"] ?: listOf(), key(ProjectileDefinition::projectileName)) })
		// tasks.add(executor.submit { loadStage(log, executor, tenants, ext2files["tenant"] ?: listOf(), key(TenantDefinition::name)) })
		tasks.add(executor.submit { loadStage(log, executor, monsterSkills, fileTree["monsterskill"] ?: listOf(), key(MonsterSkillDefinition::name)) })

		return tasks
	}

	private fun loadItemDefinitions(log: ILoadingLog, files: Map<String, Collection<IStarboundFile>>, executor: ForkJoinPool): List<ForkJoinTask<*>> {
		val fileMap = mapOf(
			"item" to ItemDefinition::class.java,
			"currency" to CurrencyItemDefinition::class.java,
			"liqitem" to LiquidItemDefinition::class.java,
			"matitem" to MaterialItemDefinition::class.java,
			"flashlight" to FlashlightDefinition::class.java,
			"harvestingtool" to HarvestingToolPrototype::class.java,
			"head" to HeadArmorItemDefinition::class.java,
			"chest" to ChestArmorItemDefinition::class.java,
			"legs" to LegsArmorItemDefinition::class.java,
			"back" to BackArmorItemDefinition::class.java,
		)

		val tasks = ArrayList<ForkJoinTask<*>>()
		val objects = Starbound.gson.getAdapter(JsonObject::class.java)

		for ((ext, clazz) in fileMap) {
			val fileList = files[ext] ?: continue
			val adapter = Starbound.gson.getAdapter(clazz)

			tasks.add(executor.submit {
				val line = log.line("Loading items '$ext'")
				line.maxElements = fileList.size
				val time = System.nanoTime()

				ParallelPerform(fileList.spliterator(), { listedFile ->
					try {
						line.text = "Loading $listedFile"
						val json = objects.read(listedFile.jsonReader())
						val def = AssetPathStack(listedFile.computeDirectory()) { adapter.fromJsonTree(json) }
						items.add(def.itemName, def, json, listedFile)
					} catch (err: Throwable) {
						LOGGER.error("Loading item definition file $listedFile", err)
					} finally {
						line.elements.incrementAndGet()
					}
				}).fork().join()

				line.text = "Loaded items '$ext' in ${((System.nanoTime() - time) / 1_000_000.0).toLong()}ms"
			})
		}

		return tasks
	}

	private fun loadJsonFunctions(line: ILoadingLog.ILine, files: Collection<IStarboundFile>) {
		for (listedFile in files) {
			line.text = ("Loading $listedFile")

			try {
				val json = Starbound.gson.getAdapter(JsonObject::class.java).read(JsonReader(listedFile.reader()).also { it.isLenient = true })

				for ((k, v) in json.entrySet()) {
					try {
						line.text = ("Loading $k from $listedFile")
						val fn = Starbound.gson.fromJson<JsonFunction>(JsonTreeReader(v), JsonFunction::class.java)
						jsonFunctions.add(k, fn, v, listedFile)
					} catch (err: Exception) {
						LOGGER.error("Loading json function definition $k from file $listedFile", err)
					}
				}
			} catch (err: Exception) {
				LOGGER.error("Loading json function definition $listedFile", err)
			}

			if (Starbound.terminateLoading) {
				return
			}
		}
	}

	private fun loadJson2Functions(line: ILoadingLog.ILine, files: Collection<IStarboundFile>) {
		for (listedFile in files) {
			line.text = ("Loading $listedFile")

			try {
				val json = Starbound.gson.getAdapter(JsonObject::class.java).read(JsonReader(listedFile.reader()).also { it.isLenient = true })

				for ((k, v) in json.entrySet()) {
					try {
						line.text = ("Loading $k from $listedFile")
						val fn = Starbound.gson.fromJson<Json2Function>(JsonTreeReader(v), Json2Function::class.java)
						json2Functions.add(k, fn, v, listedFile)
					} catch (err: Throwable) {
						LOGGER.error("Loading json 2function definition $k from file $listedFile", err)
					}
				}
			} catch (err: Exception) {
				LOGGER.error("Loading json 2function definition $listedFile", err)
			}

			if (Starbound.terminateLoading) {
				return
			}
		}
	}

	private fun loadTreasurePools(line: ILoadingLog.ILine, files: Collection<IStarboundFile>) {
		for (listedFile in files) {
			line.text = ("Loading $listedFile")

			try {
				val json = Starbound.gson.getAdapter(JsonObject::class.java).read(JsonReader(listedFile.reader()).also { it.isLenient = true })

				for ((k, v) in json.entrySet()) {
					try {
						line.text = ("Loading $k from $listedFile")
						val result = Starbound.gson.fromJson<TreasurePoolDefinition>(JsonTreeReader(v), TreasurePoolDefinition::class.java)
						result.name = k
						treasurePools.add(result.name, result, v, listedFile)
					} catch (err: Throwable) {
						LOGGER.error("Loading treasure pool definition $k from file $listedFile", err)
					}
				}
			} catch (err: Exception) {
				LOGGER.error("Loading treasure pool definition $listedFile", err)
			}

			if (Starbound.terminateLoading) {
				return
			}
		}
	}

}
