package ru.dbotthepony.kstarbound.util

import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.TypeAdapter
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import org.classdump.luna.Table
import org.classdump.luna.TableFactory
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.item.ItemDescriptor
import ru.dbotthepony.kstarbound.defs.item.api.IItemDefinition
import ru.dbotthepony.kstarbound.json.consumeNull
import ru.dbotthepony.kstarbound.lua.from

class ItemStack private constructor(item: Registry.Entry<IItemDefinition>?, count: Long, val parameters: JsonObject, marker: Unit) {
	constructor(item: Registry.Entry<IItemDefinition>, count: Long = 1L, parameters: JsonObject = JsonObject()) : this(item, count, parameters, Unit)

	var item: Registry.Entry<IItemDefinition>? = item
		private set

	var count = count
		set(value) {
			if (field == 0L || item == null)
				return

			field = value.coerceAtLeast(0L)

			if (field == 0L) {
				item = null
			}
		}

	val isEmpty: Boolean
		get() = count <= 0 || item == null

	val isNotEmpty: Boolean
		get() = count > 0 && item != null

	val maxStackSize: Long
		get() = item?.value?.maxStack ?: 0L

	fun grow(amount: Long) {
		count += amount
	}

	fun shrink(amount: Long) {
		count -= amount
	}

	fun createDescriptor(): ItemDescriptor {
		if (isEmpty)
			return ItemDescriptor("", 0L, JsonObject())

		return ItemDescriptor(item!!.key, count, parameters.deepCopy())
	}

	/**
	 * Возвращает null если этот предмет пуст
	 */
	fun conciseToNull(): ItemStack? {
		if (isEmpty) {
			return null
		} else {
			return this
		}
	}

	fun mergeFrom(other: ItemStack, simulate: Boolean) {
		if (isStackable(other)) {
			val newCount = (count + other.count).coerceAtMost(item!!.value.maxStack)
			val diff = newCount - count
			other.count -= diff

			if (!simulate)
				count = newCount
		}
	}

	fun lenientEquals(other: Any?): Boolean {
		if (other !is ItemStack)
			return false

		if (isEmpty)
			return other.isEmpty

		return other.count == count && other.item == item
	}

	fun isStackable(other: ItemStack): Boolean {
		return count != 0L && other.count != 0L && item!!.value.maxStack < count && other.item == item && other.parameters == parameters
	}

	override fun equals(other: Any?): Boolean {
		if (other !is ItemStack)
			return false

		if (isEmpty)
			return other.isEmpty

		return other.count == count && other.item == item && other.parameters == parameters
	}

	// мы не можем делать хеш из count и parameters так как они изменяемы
	override fun hashCode(): Int {
		return item.hashCode()
	}

	override fun toString(): String {
		if (isEmpty)
			return "ItemDescriptor.EMPTY"

		return "ItemDescriptor[${item!!.value.itemName}, count = $count, params = $parameters]"
	}

	fun copy(): ItemStack {
		if (isEmpty)
			return this

		return ItemStack(item, count, parameters.deepCopy(), Unit)
	}

	fun toJson(): JsonObject? {
		if (isEmpty) {
			return null
		}

		return JsonObject().also {
			it.add("name", JsonPrimitive(item!!.key))
			it.add("count", JsonPrimitive(count))
			it.add("parameters", parameters.deepCopy())
		}
	}

	fun toTable(allocator: TableFactory): Table? {
		if (isEmpty) {
			return null
		}

		return allocator.newTable(0, 3).also {
			it.rawset("name", item!!.key)
			it.rawset("count", count)
			it.rawset("parameters", allocator.from(parameters))
		}
	}

	class Adapter(val starbound: Starbound) : TypeAdapter<ItemStack>() {
		override fun write(out: JsonWriter, value: ItemStack?) {
			val json = value?.toJson()

			if (json == null)
				out.nullValue()
			else
				TypeAdapters.JSON_ELEMENT.write(out, json)
		}

		override fun read(`in`: JsonReader): ItemStack {
			if (`in`.consumeNull())
				return EMPTY

			return starbound.item(TypeAdapters.JSON_ELEMENT.read(`in`))
		}
	}

	companion object {
		@JvmField
		val EMPTY = ItemStack(null, 0L, JsonObject(), Unit)
	}
}
